import importlib
import inspect
import logging
import os.path
import pkgutil


_registered_classes = {}


def register_classes_dir_by_key(key, classes_dir: str, base_class: type, recursive: bool):
    if key in _registered_classes:
        logging.warning("Classes were registered with the same key already.")
    _registered_classes.setdefault(key, []).append((classes_dir, base_class, recursive))


def import_classes_by_key(key):
    classes = []
    if key in _registered_classes:
        registered_list = _registered_classes[key]

        for classes_dir, base_class, recursive in registered_list:
            imported_classes = import_classes_by_dir(classes_dir, base_class, recursive)
            for clas in imported_classes:
                if clas not in classes:
                    classes.append(clas)
    else:
        logging.warning(f"There are no classes registered with that key: {key}")
    return classes


def import_classes_by_dir(dir_path, base_class, recursive=True):
    """
    Importar las clases de los scripts que estén en un directorio.
    :param dir_path: directorio de los scripts.
    :param base_class: clase base de las clases que se desean importar.
    :param recursive: si se desea buscar también en los subdirectorios.
    :return:
    """
    classes = set()

    rel = os.path.relpath(dir_path)
    rel_stripped = rel.strip('./\\')
    absolute = os.path.abspath(dir_path)

    if '.' in absolute:
        logging.warning(f"Can't import modules from a directory with dot/s in it's path: {absolute}.")
        return classes

    module_dir = rel_stripped.replace('/', '.').replace('\\', '.')
    for (modinfo, name, ispkg) in pkgutil.iter_modules([dir_path]):
        module = f'{module_dir}.{name}'
        mod_classes = _import_module_classes(module, base_class)
        classes = classes.union(mod_classes)

    if recursive:
        for d in os.listdir(dir_path):
            subdir = os.path.join(dir_path, d)
            if os.path.isdir(subdir) and not subdir.endswith('__pycache__'):
                mod_classes = import_classes_by_dir(subdir, base_class, recursive)
                classes = classes.union(mod_classes)
    return classes


def _import_module_classes(module, base_class):
    """
    Importar las clases de un módulo que hereden de una clase específica.
    :param module: dirección del módulo en string. e.g. mymodule.functions
    :param base_class: clase base de las clases que se desean importar.
    :return:
    """
    imported_module = importlib.import_module(module)

    classes = set()
    for i in dir(imported_module):
        attribute = getattr(imported_module, i)
        if inspect.isclass(attribute) \
                and not inspect.isabstract(attribute) \
                and issubclass(attribute, base_class) \
                and attribute != base_class \
                and not issubclass(attribute, NotImport):
            classes.add(attribute)

    return classes


class NotImport:
    """
    Esta clase es solo para "marcar" otras clases como no importables. Es decir, las clases que hereden de NotImport
    no serán importadas de manera dinámica por el método import_module_classes.
    """

    @staticmethod
    def notimport(clazz):
        """
        Este método funciona como decorador de clases para añadir a NotImport como clase base a una clase.

        Uso:

        @notimport
        class A(object):
            pass

        :param clazz:
        :return:
        """
        if inspect.isclass(clazz):
            clazz.__bases__ += (NotImport, )
        return clazz



import setuptools
import subprocess
import sys


def get_last_version():
    bash_command = "git describe --tags --abbrev=0 --match v[0-9]*"
    process = subprocess.Popen(bash_command.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()
    outputstr = str(output)
    outputstr = outputstr[3:-3]  # "b'vXX.XX.XX\\n'" becomes "XX.XX.XX"
    outputstr = outputstr.replace('-', '.')
    major, minor, patch, *_ = outputstr.split('.')
    version = f'{major}.{minor}.{patch}'
    return version


def bump_patch(version: str):
    major, minor, patch = version.split('.')
    bumped = str(int(patch) + 1)
    return f'{major}.{minor}.{bumped}'


def push_new_version_as_tag(version):
    bash_command = f"git tag v{version}"
    process = subprocess.Popen(bash_command.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()
    bash_command = f"git push origin v{version}"
    process = subprocess.Popen(bash_command.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()


if __name__ == '__main__':
    last_version = get_last_version()
    version = bump_patch(last_version)

    if sys.argv[1] == 'pushtag':
        push_new_version_as_tag(version)
    else:
        with open("README.md", "r") as fh:
            long_description = fh.read()

        setuptools.setup(
            name="pyrulo",
            version=version,
            author="Miguel Nicolás-Díaz",
            author_email="miguelcok27@gmail.com",
            description="Python package to import modules at runtime.",
            long_description=long_description,
            long_description_content_type="text/markdown",
            url="https://github.com/mnicolas94/pyrulo",
            packages=['pyrulo'],
            classifiers=[
                "Programming Language :: Python :: 3",
                "License :: OSI Approved :: MIT License",
                "Operating System :: OS Independent",
                'Development Status :: 3 - Alpha',
                'Intended Audience :: Developers',
            ],
            python_requires='>=3.6',
        )
